@@ -33,6 +33,7 @@
 #include "filterpo.hh"
 #include "rec-snmp.hh"
 #include "rec-main.hh"
+#include "arguments.hh"
 
 boost::optional<dnsheader> RecursorLua4::DNSQuestion::getDH() const
 {
@@ -515,6 +516,13 @@ void RecursorLua4::postPrepareContext() // NOLINT(readability-function-cognitive
     thread.detach();
   });
 
+  d_lw->writeFunction("getConfigDirAndName", []() -> std::tuple<std::string, std::string> {
+      std::string dir = ::arg()["config-dir"];
+      cleanSlashes(dir);
+      std::string name = ::arg()["config-name"];
+      return {dir, name};
+  });
+
 
   if (!d_include_path.empty()) {
     includePath(d_include_path);
