@@ -72,7 +72,7 @@ private:
   string tmp;
 };
 
-void NSECBitmap::toPacket(DNSPacketWriter& pw)
+void NSECBitmap::toPacket(DNSPacketWriter& pw) const
 {
   NSECBitmapGenerator nbg(pw);
   if (d_bitset) {
@@ -108,7 +108,7 @@ void NSECBitmap::fromPacket(PacketReader& pr)
   if(bitmap.size() < 2) {
     throw MOADNSException("NSEC record with impossibly small bitmap");
   }
-  
+
   for(unsigned int n = 0; n+1 < bitmap.size();) {
     uint8_t window=static_cast<uint8_t>(bitmap[n++]);
     uint8_t blen=static_cast<uint8_t>(bitmap[n++]);
@@ -164,7 +164,7 @@ string NSECBitmap::getZoneRepresentation() const
   return ret;
 }
 
-void NSECRecordContent::report(void)
+void NSECRecordContent::report(const ReportIsOnlyCallableByReportAllTypes& /* unused */)
 {
   regist(1, 47, &make, &make, "NSEC");
 }
@@ -174,9 +174,9 @@ std::shared_ptr<DNSRecordContent> NSECRecordContent::make(const string& content)
   return std::make_shared<NSECRecordContent>(content);
 }
 
-NSECRecordContent::NSECRecordContent(const string& content, const string& zone)
+NSECRecordContent::NSECRecordContent(const string& content, const DNSName& zone)
 {
-  RecordTextReader rtr(content, DNSName(zone));
+  RecordTextReader rtr(content, zone);
   rtr.xfrName(d_next);
 
   while(!rtr.eof()) {
@@ -186,13 +186,13 @@ NSECRecordContent::NSECRecordContent(const string& content, const string& zone)
   }
 }
 
-void NSECRecordContent::toPacket(DNSPacketWriter& pw)
+void NSECRecordContent::toPacket(DNSPacketWriter& pw) const
 {
   pw.xfrName(d_next);
   d_bitmap.toPacket(pw);
 }
 
-std::shared_ptr<NSECRecordContent::DNSRecordContent> NSECRecordContent::make(const DNSRecord &dr, PacketReader& pr)
+std::shared_ptr<NSECRecordContent::DNSRecordContent> NSECRecordContent::make(const DNSRecord & /* dr */, PacketReader& pr)
 {
   auto ret=std::make_shared<NSECRecordContent>();
   pr.xfrName(ret->d_next);
@@ -202,7 +202,7 @@ std::shared_ptr<NSECRecordContent::DNSRecordContent> NSECRecordContent::make(con
   return ret;
 }
 
-string NSECRecordContent::getZoneRepresentation(bool noDot) const
+string NSECRecordContent::getZoneRepresentation(bool /* noDot */) const
 {
   string ret;
   RecordTextWriter rtw(ret);
@@ -213,7 +213,7 @@ string NSECRecordContent::getZoneRepresentation(bool noDot) const
 
 ////// begin of NSEC3
 
-void NSEC3RecordContent::report(void)
+void NSEC3RecordContent::report(const ReportIsOnlyCallableByReportAllTypes& /* unused */)
 {
   regist(1, 50, &make, &make, "NSEC3");
 }
@@ -223,16 +223,16 @@ std::shared_ptr<DNSRecordContent> NSEC3RecordContent::make(const string& content
   return std::make_shared<NSEC3RecordContent>(content);
 }
 
-NSEC3RecordContent::NSEC3RecordContent(const string& content, const string& zone)
+NSEC3RecordContent::NSEC3RecordContent(const string& content, const DNSName& zone)
 {
-  RecordTextReader rtr(content, DNSName(zone));
+  RecordTextReader rtr(content, zone);
   rtr.xfr8BitInt(d_algorithm);
   rtr.xfr8BitInt(d_flags);
   rtr.xfr16BitInt(d_iterations);
 
   rtr.xfrHexBlob(d_salt);
   rtr.xfrBase32HexBlob(d_nexthash);
-  
+
   while(!rtr.eof()) {
     uint16_t type;
     rtr.xfrType(type);
@@ -240,7 +240,7 @@ NSEC3RecordContent::NSEC3RecordContent(const string& content, const string& zone
   }
 }
 
-void NSEC3RecordContent::toPacket(DNSPacketWriter& pw) 
+void NSEC3RecordContent::toPacket(DNSPacketWriter& pw) const
 {
   pw.xfr8BitInt(d_algorithm);
   pw.xfr8BitInt(d_flags);
@@ -254,7 +254,7 @@ void NSEC3RecordContent::toPacket(DNSPacketWriter& pw)
   d_bitmap.toPacket(pw);
 }
 
-std::shared_ptr<NSEC3RecordContent::DNSRecordContent> NSEC3RecordContent::make(const DNSRecord &dr, PacketReader& pr)
+std::shared_ptr<NSEC3RecordContent::DNSRecordContent> NSEC3RecordContent::make(const DNSRecord& /* dr */, PacketReader& pr)
 {
   auto ret=std::make_shared<NSEC3RecordContent>();
   pr.xfr8BitInt(ret->d_algorithm);
@@ -271,7 +271,7 @@ std::shared_ptr<NSEC3RecordContent::DNSRecordContent> NSEC3RecordContent::make(c
   return ret;
 }
 
-string NSEC3RecordContent::getZoneRepresentation(bool noDot) const
+string NSEC3RecordContent::getZoneRepresentation(bool /* noDot */) const
 {
   string ret;
   RecordTextWriter rtw(ret);
@@ -286,7 +286,7 @@ string NSEC3RecordContent::getZoneRepresentation(bool noDot) const
 }
 
 
-void NSEC3PARAMRecordContent::report(void)
+void NSEC3PARAMRecordContent::report(const ReportIsOnlyCallableByReportAllTypes& /* unused */)
 {
   regist(1, 51, &make, &make, "NSEC3PARAM");
   regist(254, 51, &make, &make, "NSEC3PARAM");
@@ -297,45 +297,101 @@ std::shared_ptr<DNSRecordContent> NSEC3PARAMRecordContent::make(const string& co
   return std::make_shared<NSEC3PARAMRecordContent>(content);
 }
 
-NSEC3PARAMRecordContent::NSEC3PARAMRecordContent(const string& content, const string& zone) 
+NSEC3PARAMRecordContent::NSEC3PARAMRecordContent(const string& content, const DNSName& zone)
 {
-  RecordTextReader rtr(content, DNSName(zone));
-  rtr.xfr8BitInt(d_algorithm); 
-  rtr.xfr8BitInt(d_flags); 
-  rtr.xfr16BitInt(d_iterations); 
+  RecordTextReader rtr(content, zone);
+  rtr.xfr8BitInt(d_algorithm);
+  rtr.xfr8BitInt(d_flags);
+  rtr.xfr16BitInt(d_iterations);
   rtr.xfrHexBlob(d_salt);
 }
 
-void NSEC3PARAMRecordContent::toPacket(DNSPacketWriter& pw) 
+void NSEC3PARAMRecordContent::toPacket(DNSPacketWriter& pw) const
 {
-  pw.xfr8BitInt(d_algorithm); 
-        pw.xfr8BitInt(d_flags); 
-        pw.xfr16BitInt(d_iterations); 
+  pw.xfr8BitInt(d_algorithm);
+        pw.xfr8BitInt(d_flags);
+        pw.xfr16BitInt(d_iterations);
   pw.xfr8BitInt(d_salt.length());
   // cerr<<"salt: '"<<makeHexDump(d_salt)<<"', "<<d_salt.length()<<endl;
   pw.xfrBlob(d_salt);
 }
 
-std::shared_ptr<NSEC3PARAMRecordContent::DNSRecordContent> NSEC3PARAMRecordContent::make(const DNSRecord &dr, PacketReader& pr)
+std::shared_ptr<NSEC3PARAMRecordContent::DNSRecordContent> NSEC3PARAMRecordContent::make(const DNSRecord& /* dr */, PacketReader& pr)
 {
   auto ret=std::make_shared<NSEC3PARAMRecordContent>();
-  pr.xfr8BitInt(ret->d_algorithm); 
-        pr.xfr8BitInt(ret->d_flags); 
-        pr.xfr16BitInt(ret->d_iterations); 
+  pr.xfr8BitInt(ret->d_algorithm);
+        pr.xfr8BitInt(ret->d_flags);
+        pr.xfr16BitInt(ret->d_iterations);
   uint8_t len;
   pr.xfr8BitInt(len);
   pr.xfrHexBlob(ret->d_salt, len);
   return ret;
 }
 
-string NSEC3PARAMRecordContent::getZoneRepresentation(bool noDot) const
+string NSEC3PARAMRecordContent::getZoneRepresentation(bool /* noDot */) const
 {
   string ret;
   RecordTextWriter rtw(ret);
-  rtw.xfr8BitInt(d_algorithm); 
-        rtw.xfr8BitInt(d_flags); 
-        rtw.xfr16BitInt(d_iterations); 
+  rtw.xfr8BitInt(d_algorithm);
+        rtw.xfr8BitInt(d_flags);
+        rtw.xfr16BitInt(d_iterations);
   rtw.xfrHexBlob(d_salt);
   return ret;
 }
 
+////// end of NSEC3
+
+////// begin of CSYNC
+
+void CSYNCRecordContent::report(const ReportIsOnlyCallableByReportAllTypes& /* unused */)
+{
+  regist(1, 62, &make, &make, "CSYNC");
+}
+
+std::shared_ptr<DNSRecordContent> CSYNCRecordContent::make(const string& content)
+{
+  return std::make_shared<CSYNCRecordContent>(content);
+}
+
+CSYNCRecordContent::CSYNCRecordContent(const string& content, const DNSName& zone)
+{
+  RecordTextReader rtr(content, zone);
+  rtr.xfr32BitInt(d_serial);
+  rtr.xfr16BitInt(d_flags);
+
+  while(!rtr.eof()) {
+    uint16_t type;
+    rtr.xfrType(type);
+    set(type);
+  }
+}
+
+void CSYNCRecordContent::toPacket(DNSPacketWriter& pw) const
+{
+  pw.xfr32BitInt(d_serial);
+  pw.xfr16BitInt(d_flags);
+
+  d_bitmap.toPacket(pw);
+}
+
+std::shared_ptr<CSYNCRecordContent::DNSRecordContent> CSYNCRecordContent::make(const DNSRecord& /* dr */, PacketReader& pr)
+{
+  auto ret=std::make_shared<CSYNCRecordContent>();
+  pr.xfr32BitInt(ret->d_serial);
+  pr.xfr16BitInt(ret->d_flags);
+
+  ret->d_bitmap.fromPacket(pr);
+  return ret;
+}
+
+string CSYNCRecordContent::getZoneRepresentation(bool /* noDot */) const
+{
+  string ret;
+  RecordTextWriter rtw(ret);
+  rtw.xfr32BitInt(d_serial);
+  rtw.xfr16BitInt(d_flags);
+
+  return ret + d_bitmap.getZoneRepresentation();
+}
+
+////// end of CSYNC
