@@ -247,6 +247,10 @@ static RecursorControlChannel::Answer showYAML(const std::string& path)
   try {
     std::string msg;
     auto converted = pdns::settings::rec::oldStyleSettingsFileToYaml(configName, true);
+    if (converted == "{}\n") {
+      msg += "There seems to be no YAML config in " + configName;
+      return {1, std::move(msg)};
+    }
     msg += "# Start of converted recursor.yml based on " + configName + "\n";
     msg += converted;
     msg += "# Validation result: ";
