@@ -24,16 +24,15 @@
 #include "dnsparser.hh"
 #include "dnsrecords.hh"
 #include "ixfr.hh"
-#include "syncres.hh"
 #include "axfr-retriever.hh"
 #include "lock.hh"
-#include "logger.hh"
 #include "logging.hh"
 #include "rec-lua-conf.hh"
 #include "rpzloader.hh"
 #include "zoneparser-tng.hh"
 #include "threadname.hh"
 #include "query-local-address.hh"
+#include "rec-system-resolve.hh"
 
 Netmask makeNetmaskFromRPZ(const DNSName& name)
 {
@@ -383,6 +382,19 @@ std::shared_ptr<const SOARecordContent> loadRPZFromFile(const std::string& fname
 static bool dumpZoneToDisk(Logr::log_t logger, const std::shared_ptr<DNSFilterEngine::Zone>& newZone, const std::string& dumpZoneFileName)
 {
   logger->info(Logr::Debug, "Dumping zone to disk", "destination_file", Logging::Loggable(dumpZoneFileName));
+  DNSRecord soa = newZone->getSOA();
+  uint32_t serial = 0;
+  DNSName zone;
+  if (auto soaContent = getRR<SOARecordContent>(soa)) {
+    serial = soaContent->d_st.serial;
+  }
+  if (newZone->getSerial() != serial) {
+    logger->info(Logr::Error, "Inconsistency of internal serial and SOA serial", "serial", Logging::Loggable(newZone->getSerial()), "soaserial", Logging::Loggable(serial));
+  }
+
+  if (newZone->getDomain() != soa.d_name) {
+    logger->info(Logr::Error, "Inconsistency of internal name and SOA name", "zone", Logging::Loggable(newZone->getDomain()), "soaname", Logging::Loggable(soa.d_name));
+  }
   std::string temp = dumpZoneFileName + "XXXXXX";
   int fileDesc = mkstemp(&temp.at(0));
   if (fileDesc < 0) {
@@ -436,34 +448,23 @@ static bool dumpZoneToDisk(Logr::log_t logger, const std::shared_ptr<DNSFilterEn
   return true;
 }
 
-// A struct that holds the condition var and related stuff to allow notifies to be sent to the tread owning
-// the struct.
-struct RPZWaiter
+static void preloadRPZFIle(RPZTrackerParams& params, const DNSName& zoneName, std::shared_ptr<DNSFilterEngine::Zone>& oldZone, uint32_t& refresh, const string& polName, uint64_t configGeneration, ZoneXFR::ZoneWaiter& rpzwaiter, Logr::log_t logger)
 {
-  RPZWaiter(std::thread::id arg) :
-    id(arg) {}
-  std::thread::id id;
-  std::mutex mutex;
-  std::condition_variable condVar;
-  std::atomic<bool> stop{false};
-};
-
-static void preloadRPZFIle(RPZTrackerParams& params, const DNSName& zoneName, std::shared_ptr<DNSFilterEngine::Zone>& oldZone, uint32_t& refresh, const string& polName, uint64_t configGeneration, RPZWaiter& rpzwaiter, Logr::log_t logger)
-{
-  while (!params.soaRecordContent) {
+  while (!params.zoneXFRParams.soaRecordContent) {
     /* if we received an empty sr, the zone was not really preloaded */
 
     /* full copy, as promised */
     std::shared_ptr<DNSFilterEngine::Zone> newZone = std::make_shared<DNSFilterEngine::Zone>(*oldZone);
-    for (const auto& primary : params.primaries) {
+    for (const auto& primary : params.zoneXFRParams.primaries) {
       try {
-        params.soaRecordContent = loadRPZFromServer(logger, primary, zoneName, newZone, params.defpol, params.defpolOverrideLocal, params.maxTTL, params.tsigtriplet, params.maxReceivedMBytes, params.localAddress, params.xfrTimeout);
-        newZone->setSerial(params.soaRecordContent->d_st.serial);
-        newZone->setRefresh(params.soaRecordContent->d_st.refresh);
-        refresh = std::max(params.refreshFromConf != 0 ? params.refreshFromConf : newZone->getRefresh(), 1U);
-        setRPZZoneNewState(polName, params.soaRecordContent->d_st.serial, newZone->size(), false, true);
+        auto combo = pdns::fromNameOrIP(primary, 53, logger);
+        params.zoneXFRParams.soaRecordContent = loadRPZFromServer(logger, combo, zoneName, newZone, params.defpol, params.defpolOverrideLocal, params.maxTTL, params.zoneXFRParams.tsigtriplet, params.zoneXFRParams.maxReceivedMBytes, params.zoneXFRParams.localAddress, params.zoneXFRParams.xfrTimeout);
+        newZone->setSerial(params.zoneXFRParams.soaRecordContent->d_st.serial);
+        newZone->setRefresh(params.zoneXFRParams.soaRecordContent->d_st.refresh);
+        refresh = std::max(params.zoneXFRParams.refreshFromConf != 0 ? params.zoneXFRParams.refreshFromConf : newZone->getRefresh(), 1U);
+        setRPZZoneNewState(polName, params.zoneXFRParams.soaRecordContent->d_st.serial, newZone->size(), false, true);
 
-        g_luaconfs.modify([zoneIdx = params.zoneIdx, &newZone](LuaConfigItems& lci) {
+        g_luaconfs.modify([zoneIdx = params.zoneXFRParams.zoneIdx, &newZone](LuaConfigItems& lci) {
           lci.dfe.setZone(zoneIdx, newZone);
         });
 
@@ -487,7 +488,7 @@ static void preloadRPZFIle(RPZTrackerParams& params, const DNSName& zoneName, st
     }
     // Release newZone before (long) sleep to reduce memory usage
     newZone = nullptr;
-    if (!params.soaRecordContent) {
+    if (!params.zoneXFRParams.soaRecordContent) {
       std::unique_lock lock(rpzwaiter.mutex);
       rpzwaiter.condVar.wait_for(lock, std::chrono::seconds(refresh),
                                  [&stop = rpzwaiter.stop] { return stop.load(); });
@@ -504,12 +505,12 @@ static void preloadRPZFIle(RPZTrackerParams& params, const DNSName& zoneName, st
   }
 }
 
-static bool RPZTrackerIteration(RPZTrackerParams& params, const DNSName& zoneName, std::shared_ptr<DNSFilterEngine::Zone>& oldZone, uint32_t& refresh, const string& polName, bool& skipRefreshDelay, uint64_t configGeneration, RPZWaiter& rpzwaiter, Logr::log_t logger)
+static bool RPZTrackerIteration(RPZTrackerParams& params, const DNSName& zoneName, std::shared_ptr<DNSFilterEngine::Zone>& oldZone, uint32_t& refresh, const string& polName, bool& skipRefreshDelay, uint64_t configGeneration, ZoneXFR::ZoneWaiter& rpzwaiter, Logr::log_t logger)
 {
   // Don't hold on to oldZone, it well be re-assigned after sleep in the try block
   oldZone = nullptr;
   DNSRecord dnsRecord;
-  dnsRecord.setContent(params.soaRecordContent);
+  dnsRecord.setContent(params.zoneXFRParams.soaRecordContent);
 
   if (skipRefreshDelay) {
     skipRefreshDelay = false;
@@ -542,19 +543,20 @@ static bool RPZTrackerIteration(RPZTrackerParams& params, const DNSName& zoneNam
   }
 
   vector<pair<vector<DNSRecord>, vector<DNSRecord>>> deltas;
-  for (const auto& primary : params.primaries) {
+  for (const auto& ipOrName : params.zoneXFRParams.primaries) {
+    auto primary = pdns::fromNameOrIP(ipOrName, 53, logger);
     auto soa = getRR<SOARecordContent>(dnsRecord);
     auto serial = soa ? soa->d_st.serial : 0;
     SLOG(g_log << Logger::Info << "Getting IXFR deltas for " << zoneName << " from " << primary.toStringWithPort() << ", our serial: " << serial << endl,
          logger->info(Logr::Info, "Getting IXFR deltas", "address", Logging::Loggable(primary), "ourserial", Logging::Loggable(serial)));
 
-    ComboAddress local(params.localAddress);
+    ComboAddress local(params.zoneXFRParams.localAddress);
     if (local == ComboAddress()) {
       local = pdns::getQueryLocalAddress(primary.sin4.sin_family, 0);
     }
 
     try {
-      deltas = getIXFRDeltas(primary, zoneName, dnsRecord, params.xfrTimeout, true, params.tsigtriplet, &local, params.maxReceivedMBytes);
+      deltas = getIXFRDeltas(primary, zoneName, dnsRecord, params.zoneXFRParams.xfrTimeout, true, params.zoneXFRParams.tsigtriplet, &local, params.zoneXFRParams.maxReceivedMBytes);
 
       /* no need to try another primary */
       break;
@@ -580,7 +582,7 @@ static bool RPZTrackerIteration(RPZTrackerParams& params, const DNSName& zoneNam
            logger->info(Logr::Info, "A more recent configuration has been found, stopping the existing RPZ update thread"));
       return false;
     }
-    oldZone = luaconfsLocal->dfe.getZone(params.zoneIdx);
+    oldZone = luaconfsLocal->dfe.getZone(params.zoneXFRParams.zoneIdx);
     if (!oldZone || oldZone->getDomain() != zoneName) {
       SLOG(g_log << Logger::Info << "This policy is no more, stopping the existing RPZ update thread for " << zoneName << endl,
            logger->info(Logr::Info, "This policy is no more, stopping the existing RPZ update thread"));
@@ -589,7 +591,7 @@ static bool RPZTrackerIteration(RPZTrackerParams& params, const DNSName& zoneNam
     /* we need to make a _full copy_ of the zone we are going to work on */
     std::shared_ptr<DNSFilterEngine::Zone> newZone = std::make_shared<DNSFilterEngine::Zone>(*oldZone);
     /* initialize the current serial to the last one */
-    std::shared_ptr<const SOARecordContent> currentSR = params.soaRecordContent;
+    std::shared_ptr<const SOARecordContent> currentSR = params.zoneXFRParams.soaRecordContent;
 
     int totremove = 0;
     int totadd = 0;
@@ -632,8 +634,10 @@ static bool RPZTrackerIteration(RPZTrackerParams& params, const DNSName& zoneNam
           continue;
         }
         if (resourceRecord.d_type == QType::SOA) {
-          auto tempSR = getRR<SOARecordContent>(resourceRecord);
-          if (tempSR) {
+          if (auto tempSR = getRR<SOARecordContent>(resourceRecord)) {
+            dnsRecord = resourceRecord;
+            // IXFR leaves us a relative name, fix that
+            dnsRecord.d_name = newZone->getDomain();
             currentSR = std::move(tempSR);
           }
         }
@@ -648,13 +652,14 @@ static bool RPZTrackerIteration(RPZTrackerParams& params, const DNSName& zoneNam
 
     /* only update sr now that all changes have been converted */
     if (currentSR) {
-      params.soaRecordContent = std::move(currentSR);
+      newZone->setSOA(std::move(dnsRecord));
+      params.zoneXFRParams.soaRecordContent = std::move(currentSR);
     }
     SLOG(g_log << Logger::Info << "Had " << totremove << " RPZ removal" << addS(totremove) << ", " << totadd << " addition" << addS(totadd) << " for " << zoneName << " New serial: " << params.soaRecordContent->d_st.serial << endl,
-         logger->info(Logr::Info, "RPZ mutations", "removals", Logging::Loggable(totremove), "additions", Logging::Loggable(totadd), "newserial", Logging::Loggable(params.soaRecordContent->d_st.serial)));
-    newZone->setSerial(params.soaRecordContent->d_st.serial);
-    newZone->setRefresh(params.soaRecordContent->d_st.refresh);
-    setRPZZoneNewState(polName, params.soaRecordContent->d_st.serial, newZone->size(), false, fullUpdate);
+         logger->info(Logr::Info, "RPZ mutations", "removals", Logging::Loggable(totremove), "additions", Logging::Loggable(totadd), "newserial", Logging::Loggable(params.zoneXFRParams.soaRecordContent->d_st.serial)));
+    newZone->setSerial(params.zoneXFRParams.soaRecordContent->d_st.serial);
+    newZone->setRefresh(params.zoneXFRParams.soaRecordContent->d_st.refresh);
+    setRPZZoneNewState(polName, params.zoneXFRParams.soaRecordContent->d_st.serial, newZone->size(), false, fullUpdate);
 
     /* we need to replace the existing zone with the new one,
        but we don't want to touch anything else, especially other zones,
@@ -665,14 +670,14 @@ static bool RPZTrackerIteration(RPZTrackerParams& params, const DNSName& zoneNam
            logger->info(Logr::Info, "A more recent configuration has been found, stopping the existing RPZ update thread"));
       return false;
     }
-    g_luaconfs.modify([zoneIdx = params.zoneIdx, &newZone](LuaConfigItems& lci) {
+    g_luaconfs.modify([zoneIdx = params.zoneXFRParams.zoneIdx, &newZone](LuaConfigItems& lci) {
       lci.dfe.setZone(zoneIdx, newZone);
     });
 
     if (!params.dumpZoneFileName.empty()) {
       dumpZoneToDisk(logger, newZone, params.dumpZoneFileName);
     }
-    refresh = std::max(params.refreshFromConf != 0 ? params.refreshFromConf : newZone->getRefresh(), 1U);
+    refresh = std::max(params.zoneXFRParams.refreshFromConf != 0 ? params.zoneXFRParams.refreshFromConf : newZone->getRefresh(), 1U);
   }
   catch (const std::exception& e) {
     SLOG(g_log << Logger::Error << "Error while applying the update received over XFR for " << zoneName << ", skipping the update: " << e.what() << endl,
@@ -685,56 +690,32 @@ static bool RPZTrackerIteration(RPZTrackerParams& params, const DNSName& zoneNam
   return true;
 }
 
-// As there can be multiple threads doing updates (due to config reloads), we use a multimap.
-// The value contains the actual thread id that owns the struct.
-
-static LockGuarded<std::multimap<DNSName, RPZWaiter&>> condVars;
-
-// Notify all threads tracking the RPZ name
-bool notifyRPZTracker(const DNSName& name)
-{
-  auto lock = condVars.lock();
-  auto [start, end] = lock->equal_range(name);
-  if (start == end) {
-    // Did not find any thread tracking that RPZ name
-    return false;
-  }
-  while (start != end) {
-    start->second.stop = true;
-    start->second.condVar.notify_one();
-    ++start;
-  }
-  return true;
-}
-
 // coverity[pass_by_value] params is intended to be a copy, as this is the main function of a thread
 void RPZIXFRTracker(RPZTrackerParams params, uint64_t configGeneration)
 {
   setThreadName("rec/rpzixfr");
-  bool isPreloaded = params.soaRecordContent != nullptr;
+  bool isPreloaded = params.zoneXFRParams.soaRecordContent != nullptr;
   auto logger = g_slog->withName("rpz");
-  RPZWaiter waiter(std::this_thread::get_id());
+  ZoneXFR::ZoneWaiter waiter(std::this_thread::get_id());
 
   /* we can _never_ modify this zone directly, we need to do a full copy then replace the existing zone */
-  std::shared_ptr<DNSFilterEngine::Zone> oldZone = g_luaconfs.getLocal()->dfe.getZone(params.zoneIdx);
+  std::shared_ptr<DNSFilterEngine::Zone> oldZone = g_luaconfs.getLocal()->dfe.getZone(params.zoneXFRParams.zoneIdx);
   if (!oldZone) {
     SLOG(g_log << Logger::Error << "Unable to retrieve RPZ zone with index " << params.zoneIdx << " from the configuration, exiting" << endl,
-         logger->error(Logr::Error, "Unable to retrieve RPZ zone from configuration", "index", Logging::Loggable(params.zoneIdx)));
+         logger->error(Logr::Error, "Unable to retrieve RPZ zone from configuration", "index", Logging::Loggable(params.zoneXFRParams.zoneIdx)));
     return;
   }
 
   // If oldZone failed to load its getRefresh() returns 0, protect against that
-  uint32_t refresh = std::max(params.refreshFromConf != 0 ? params.refreshFromConf : oldZone->getRefresh(), 10U);
+  uint32_t refresh = std::max(params.zoneXFRParams.refreshFromConf != 0 ? params.zoneXFRParams.refreshFromConf : oldZone->getRefresh(), 10U);
   DNSName zoneName = oldZone->getDomain();
   std::string polName = !oldZone->getName().empty() ? oldZone->getName() : zoneName.toStringNoDot();
 
   // Now that we know the name, set it in the logger
   logger = logger->withValues("zone", Logging::Loggable(zoneName));
 
-  {
-    auto lock = condVars.lock();
-    lock->emplace(zoneName, waiter);
-  }
+  ZoneXFR::insertZoneTracker(zoneName, waiter);
+
   preloadRPZFIle(params, zoneName, oldZone, refresh, polName, configGeneration, waiter, logger);
 
   bool skipRefreshDelay = isPreloaded;
@@ -743,14 +724,5 @@ void RPZIXFRTracker(RPZTrackerParams params, uint64_t configGeneration)
     // empty
   }
 
-  // Zap our (and only our) RPZWaiter struct out of the multimap
-  auto lock = condVars.lock();
-  auto [start, end] = lock->equal_range(zoneName);
-  while (start != end) {
-    if (start->second.id == std::this_thread::get_id()) {
-      lock->erase(start);
-      break;
-    }
-    ++start;
-  }
+  ZoneXFR::clearZoneTracker(zoneName);
 }
